@@ -1,5 +1,6 @@
 /*
  * Copyright (c) [2011-2015] Novell, Inc.
+ * Copyright (c) [2016-2018] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -30,9 +31,6 @@
 #include <fcntl.h>
 #include <sys/ioctl.h>
 #include <asm/types.h>
-#ifdef ENABLE_ROLLBACK
-#include <libmount/libmount.h>
-#endif
 #ifdef HAVE_LIBBTRFS
 #ifdef HAVE_BTRFS_VERSION_H
 #include <btrfs/version.h>
@@ -56,6 +54,14 @@
 #include "snapper/SnapperTmpl.h"
 #include "snapper/SnapperDefines.h"
 #include "snapper/Acls.h"
+#include "snapper/Exception.h"
+#include "snapper/Regex.h"
+#ifdef ENABLE_ROLLBACK
+#include "snapper/MntTable.h"
+#endif
+#ifdef ENABLE_SELINUX
+#include "snapper/Selinux.h"
+#endif
 
 
 namespace snapper
@@ -82,19 +88,27 @@ namespace snapper
     void
     Btrfs::evalConfigInfo(const ConfigInfo& config_info)
     {
+#ifdef ENABLE_BTRFS_QUOTA
+
 	string qgroup_str;
 	if (config_info.getValue("QGROUP", qgroup_str) && !qgroup_str.empty())
 	{
 	    try
 	    {
-		qgroup = make_qgroup(qgroup_str);
+		qgroup = parse_qgroup(qgroup_str);
 	    }
 	    catch (const runtime_error& e)
 	    {
 		y2err("failed to parse qgroup '" << qgroup_str << "'");
-		throw InvalidConfigException();
+		SN_THROW(InvalidConfigException());
 	    }
 	}
+	else
+	{
+	    qgroup = no_qgroup;
+	}
+
+#endif
     }
 
 
@@ -107,13 +121,35 @@ namespace snapper
 	{
 	    create_subvolume(subvolume_dir.fd(), ".snapshots");
 	}
-	catch (const runtime_error& e)
+	catch (const runtime_error_with_errno& e)
 	{
 	    y2err("create subvolume failed, " << e.what());
-	    throw CreateConfigFailedException("creating btrfs snapshot failed");
+
+	    switch (e.error_number)
+	    {
+		case EEXIST:
+		    SN_THROW(CreateConfigFailedException("creating btrfs subvolume .snapshots failed "
+							 "since it already exists"));
+
+		default:
+		    SN_THROW(CreateConfigFailedException("creating btrfs subvolume .snapshots failed"));
+	    }
 	}
 
 	SFile x(subvolume_dir, ".snapshots");
+#ifdef ENABLE_SELINUX
+	try
+	{
+	    SnapperContexts scontexts;
+
+	    x.fsetfilecon(scontexts.subvolume_context());
+	}
+	catch (const SelinuxException& e)
+	{
+	    SN_CAUGHT(e);
+	    // fall through intentional
+	}
+#endif
 	struct stat stat;
 	if (x.stat(&stat, 0) == 0)
 	    x.chmod(stat.st_mode & ~0027, 0);
@@ -141,7 +177,7 @@ namespace snapper
 	catch (const runtime_error& e)
 	{
 	    y2err("delete subvolume failed, " << e.what());
-	    throw DeleteConfigFailedException("deleting btrfs snapshot failed");
+	    SN_THROW(DeleteConfigFailedException("deleting btrfs snapshot failed"));
 	}
     }
 
@@ -200,13 +236,12 @@ namespace snapper
 	struct stat stat;
 	if (subvolume_dir.stat(&stat) != 0)
 	{
-	    throw IOErrorException();
+	    SN_THROW(IOErrorException("stat on subvolume directory failed"));
 	}
 
 	if (!is_subvolume(stat))
 	{
-	    y2err("subvolume is not a btrfs snapshot");
-	    throw IOErrorException();
+	    SN_THROW(IOErrorException("subvolume is not a btrfs subvolume"));
 	}
 
 	return subvolume_dir;
@@ -222,31 +257,30 @@ namespace snapper
 	struct stat stat;
 	if (infos_dir.stat(&stat) != 0)
 	{
-	    throw IOErrorException();
+	    SN_THROW(IOErrorException("stat on info directory failed"));
 	}
 
 	if (!is_subvolume(stat))
 	{
-	    y2err(".snapshots is not a btrfs snapshot");
-	    throw IOErrorException();
+	    SN_THROW(IOErrorException(".snapshots is not a btrfs subvolume"));
 	}
 
 	if (stat.st_uid != 0)
 	{
 	    y2err(".snapshots must have owner root");
-	    throw IOErrorException();
+	    SN_THROW(IOErrorException(".snapshots must have owner root"));
 	}
 
 	if (stat.st_gid != 0 && stat.st_mode & S_IWGRP)
 	{
 	    y2err(".snapshots must have group root or must not be group-writable");
-	    throw IOErrorException();
+	    SN_THROW(IOErrorException(".snapshots must have group root or must not be group-writable"));
 	}
 
 	if (stat.st_mode & S_IWOTH)
 	{
 	    y2err(".snapshots must not be world-writable");
-	    throw IOErrorException();
+	    SN_THROW(IOErrorException(".snapshots must not be world-writable"));
 	}
 
 	return infos_dir;
@@ -263,8 +297,16 @@ namespace snapper
     }
 
 
+    SDir
+    Btrfs::openGeneralDir() const
+    {
+	return openInfosDir();
+    }
+
+
     void
-    Btrfs::createSnapshot(unsigned int num, unsigned int num_parent, bool read_only) const
+    Btrfs::createSnapshot(unsigned int num, unsigned int num_parent, bool read_only, bool quota,
+			  bool empty) const
     {
 	if (num_parent == 0)
 	{
@@ -273,12 +315,16 @@ namespace snapper
 
 	    try
 	    {
-		create_snapshot(subvolume_dir.fd(), info_dir.fd(), "snapshot", read_only, qgroup);
+		if (empty)
+		    create_subvolume(info_dir.fd(), "snapshot");
+		else
+		    create_snapshot(subvolume_dir.fd(), info_dir.fd(), "snapshot", read_only,
+				    quota ? qgroup : no_qgroup);
 	    }
 	    catch (const runtime_error& e)
 	    {
 		y2err("create snapshot failed, " << e.what());
-		throw CreateSnapshotFailedException();
+		SN_THROW(CreateSnapshotFailedException());
 	    }
 	}
 	else
@@ -288,12 +334,13 @@ namespace snapper
 
 	    try
 	    {
-		create_snapshot(snapshot_dir.fd(), info_dir.fd(), "snapshot", read_only, qgroup);
+		create_snapshot(snapshot_dir.fd(), info_dir.fd(), "snapshot", read_only,
+				quota ? qgroup : no_qgroup);
 	    }
 	    catch (const runtime_error& e)
 	    {
 		y2err("create snapshot failed, " << e.what());
-		throw CreateSnapshotFailedException();
+		SN_THROW(CreateSnapshotFailedException());
 	    }
 	}
     }
@@ -302,10 +349,10 @@ namespace snapper
 #ifdef ENABLE_ROLLBACK
 
     void
-    Btrfs::createSnapshotOfDefault(unsigned int num, bool read_only) const
+    Btrfs::createSnapshotOfDefault(unsigned int num, bool read_only, bool quota) const
     {
 	SDir subvolume_dir = openSubvolumeDir();
-	unsigned long long id = get_default_id(subvolume_dir.fd());
+	subvolid_t id = get_default_id(subvolume_dir.fd());
 	string name = get_subvolume(subvolume_dir.fd(), id);
 
 	bool found = false;
@@ -313,7 +360,7 @@ namespace snapper
 	if (!getMtabData(subvolume, found, mtab_data))
 	{
 	    y2err("failed to find device");
-	    throw CreateSnapshotFailedException();
+	    SN_THROW(CreateSnapshotFailedException());
 	}
 
 	SDir infos_dir = openInfosDir();
@@ -325,19 +372,20 @@ namespace snapper
 
 	try
 	{
-	    create_snapshot(tmp_mount_dir.fd(), info_dir.fd(), "snapshot", read_only, qgroup);
+	    create_snapshot(tmp_mount_dir.fd(), info_dir.fd(), "snapshot", read_only,
+			    quota ? qgroup : no_qgroup);
 	}
 	catch (const runtime_error& e)
 	{
 	    y2err("create snapshot failed, " << e.what());
-	    throw CreateSnapshotFailedException();
+	    SN_THROW(CreateSnapshotFailedException());
 	}
     }
 
 #else
 
     void
-    Btrfs::createSnapshotOfDefault(unsigned int num, bool read_only) const
+    Btrfs::createSnapshotOfDefault(unsigned int num, bool read_only, bool quota) const
     {
 	throw std::logic_error("not implemented");
     }
@@ -352,12 +400,39 @@ namespace snapper
 
 	try
 	{
+#ifdef HAVE_LIBBTRFS
+	    subvolid_t subvolid = get_id(openSnapshotDir(num).fd());
+#endif
+
 	    delete_subvolume(info_dir.fd(), "snapshot");
+
+#ifdef HAVE_LIBBTRFS
+	    deleted_subvolids.push_back(subvolid);
+#endif
+
+#ifdef ENABLE_BTRFS_QUOTA
+
+	    // workaround for the kernel not deleting the qgroup of a
+	    // subvolume when deleting the subvolume, see
+	    // https://bugzilla.suse.com/show_bug.cgi?id=972511
+
+	    try
+	    {
+		SDir subvolume_dir = openSubvolumeDir();
+		qgroup_destroy(subvolume_dir.fd(), calc_qgroup(0, subvolid));
+	    }
+	    catch (const runtime_error& e)
+	    {
+		// Ignore that the qgroup could not be destroyed. Should not
+		// cause problems except of having unused qgroups.
+	    }
+
+#endif
 	}
 	catch (const runtime_error& e)
 	{
 	    y2err("delete snapshot failed, " << e.what());
-	    throw DeleteSnapshotFailedException();
+	    SN_THROW(DeleteSnapshotFailedException());
 	}
     }
 
@@ -402,6 +477,8 @@ namespace snapper
 	}
 	catch (const IOErrorException& e)
 	{
+	    // TODO the openInfoDir above logs an error although when this
+	    // function is used from nextNumber the failure is ok
 	    return false;
 	}
     }
@@ -568,10 +645,9 @@ namespace snapper
     }
 
 
-    struct BtrfsSendReceiveException : public SnapperException
+    struct BtrfsSendReceiveException : public Exception
     {
-	explicit BtrfsSendReceiveException() throw() {}
-	virtual const char* what() const throw() { return "Btrfs send/receive error"; }
+	explicit BtrfsSendReceiveException() : Exception("btrfs send/receive error") {}
     };
 
 
@@ -681,7 +757,7 @@ namespace snapper
 	if (r < 0)
 	{
 	    y2err("failed to initialize subvol search (" << stringerror(r) << ")");
-	    throw BtrfsSendReceiveException();
+	    SN_THROW(BtrfsSendReceiveException());
 	}
     }
 
@@ -1204,7 +1280,7 @@ namespace snapper
 	if (r1 < 0)
 	{
 	    y2err("pipe failed errno:" << errno << " (" << stringerror(errno) << ")");
-	    throw BtrfsSendReceiveException();
+	    SN_THROW(BtrfsSendReceiveException());
 	}
 
 	struct btrfs_ioctl_send_args io_send;
@@ -1236,7 +1312,7 @@ namespace snapper
 
 	if (r2 < 0 || !uf.get())
 	{
-	    throw BtrfsSendReceiveException();
+	    SN_THROW(BtrfsSendReceiveException());
 	}
 
 #else
@@ -1257,7 +1333,7 @@ namespace snapper
 
 	if (r2 < 0 || !dumper_ret)
 	{
-	    throw BtrfsSendReceiveException();
+	    SN_THROW(BtrfsSendReceiveException());
 	}
 
 #endif
@@ -1270,7 +1346,7 @@ namespace snapper
 	u64 flags;
 	if (ioctl(dir.fd(), BTRFS_IOC_SUBVOL_GETFLAGS, &flags) < 0)
 	{
-	    throw IOErrorException();
+	    SN_THROW(IOErrorException("ioctl BTRFS_IOC_SUBVOL_GETFLAGS failed"));
 	}
 
 	return flags & BTRFS_SUBVOL_RDONLY;
@@ -1285,7 +1361,7 @@ namespace snapper
 	if (!is_subvolume_ro(dir1) || !is_subvolume_ro(dir2))
 	{
 	    y2err("not read-only snapshots");
-	    throw BtrfsSendReceiveException();
+	    SN_THROW(BtrfsSendReceiveException());
 	}
 
 	u64 parent_root_id = 0;
@@ -1293,7 +1369,7 @@ namespace snapper
 	if (!get_root_id(name1, &parent_root_id))
 	{
 	    y2err("could not resolve root_id for " << name1);
-	    throw BtrfsSendReceiveException();
+	    SN_THROW(BtrfsSendReceiveException());
 	}
 
 	vector<u64> clone_sources;
@@ -1323,7 +1399,7 @@ namespace snapper
 
 	    y2mil("stopwatch " << stopwatch << " for comparing directories");
 	}
-	catch (const SnapperException& e)
+	catch (const Exception& e)
 	{
 	    y2err("special btrfs cmpDirs failed, " << e.what());
 	    y2mil("cmpDirs fallback");
@@ -1349,107 +1425,189 @@ namespace snapper
 
 #ifdef ENABLE_ROLLBACK
 
+    bool
+    Btrfs::isDefault(unsigned int num) const
+    {
+	bool ret = false;
+
+	try
+	{
+	    SDir subvolume_dir = openSubvolumeDir();
+	    subvolid_t id = get_default_id(subvolume_dir.fd());
+	    if (num == 0)
+	    {
+		ret = get_id(subvolume_dir.fd()) == id;
+	    }
+	    else
+	    {
+		ret = get_id(openSnapshotDir(num).fd()) == id;
+	    }
+	}
+	catch (const runtime_error& e)
+	{
+	    SN_THROW(IOErrorException(string("get default failed, ") + e.what()));
+	}
+
+	return ret;
+    }
+
+
+    std::pair<bool, unsigned int>
+    Btrfs::idToNum(int fd, subvolid_t id) const
+    {
+	string path = get_subvolume(fd, id);
+
+	Regex rx("/([0-9]+)/snapshot$");
+	if (!rx.match(path))
+	    return make_pair(false, 0);
+
+	unsigned int num = stoi(rx.cap(1));
+
+	if (get_id(openSnapshotDir(num).fd()) != id)
+	    return make_pair(false, 0);
+
+	return make_pair(true, num);
+    }
+
+
+    std::pair<bool, unsigned int>
+    Btrfs::getDefault() const
+    {
+	SDir subvolume_dir = openSubvolumeDir();
+	int fd = subvolume_dir.fd();
+
+	return idToNum(fd, get_default_id(fd));
+    }
+
+
     void
     Btrfs::setDefault(unsigned int num) const
     {
 	try
 	{
+	    SDir general_dir = openGeneralDir();
+
 	    if (num == 0)
 	    {
 		SDir subvolume_dir = openSubvolumeDir();
-		unsigned long long id = get_id(subvolume_dir.fd());
-		set_default_id(subvolume_dir.fd(), id);
+		subvolid_t id = get_id(subvolume_dir.fd());
+		set_default_id(general_dir.fd(), id);
 	    }
 	    else
 	    {
 		SDir snapshot_dir = openSnapshotDir(num);
-		unsigned long long id = get_id(snapshot_dir.fd());
-
-		SDir subvolume_dir = openSubvolumeDir();
-		set_default_id(subvolume_dir.fd(), id);
+		subvolid_t id = get_id(snapshot_dir.fd());
+		set_default_id(general_dir.fd(), id);
 	    }
 	}
 	catch (const runtime_error& e)
 	{
-	    y2err("set default failed, " << e.what());
-	    throw IOErrorException();
+	    SN_THROW(IOErrorException(string("set default failed, ") + e.what()));
 	}
     }
 
+
+    std::pair<bool, unsigned int>
+    Btrfs::getActive() const
+    {
+	SDir subvolume_dir = openSubvolumeDir();
+	int fd = subvolume_dir.fd();
+
+	return idToNum(fd, get_id(fd));
+    }
+
+
+    bool
+    Btrfs::isActive(unsigned int num) const
+    {
+	bool ret = false;
+
+	try
+	{
+	    if (num == 0)
+		SN_THROW(IllegalSnapshotException());
+
+	    SDir snapshot_dir = openSnapshotDir(num);
+	    SDir subvolume_dir = openSubvolumeDir();
+	    ret = get_id(snapshot_dir.fd()) == get_id(subvolume_dir.fd());
+	}
+	catch (const runtime_error& e)
+	{
+	    SN_THROW(IOErrorException(string("get active failed, ") + e.what()));
+	}
+
+	return ret;
+    }
+
 #else
 
+    bool
+    Btrfs::isDefault(unsigned int num) const
+    {
+	SN_THROW(UnsupportedException());
+	__builtin_unreachable();
+    }
+
+
+    std::pair<bool, unsigned int>
+    Btrfs::getDefault() const
+    {
+	SN_THROW(UnsupportedException());
+	__builtin_unreachable();
+    }
+
+
     void
     Btrfs::setDefault(unsigned int num) const
     {
-	throw std::logic_error("not implemented");
+	SN_THROW(UnsupportedException());
+	__builtin_unreachable();
+    }
+
+
+    std::pair<bool, unsigned int>
+    Btrfs::getActive() const
+    {
+	SN_THROW(UnsupportedException());
+	__builtin_unreachable();
+    }
+
+
+    bool
+    Btrfs::isActive(unsigned int num) const
+    {
+	SN_THROW(UnsupportedException());
+	__builtin_unreachable();
     }
 
 #endif
 
 
-#ifdef ENABLE_ROLLBACK
-
-    class MntTable
+    void
+    Btrfs::sync() const
     {
+	SDir subvolume_dir = openSubvolumeDir();
 
-    public:
+	BtrfsUtils::sync(subvolume_dir.fd());
 
-	MntTable(const string& root_prefix)
-	    : root_prefix(root_prefix), table(mnt_new_table())
+	if (!deleted_subvolids.empty())
 	{
-	    if (!table)
-		throw runtime_error("mnt_new_table failed");
+#ifdef HAVE_LIBBTRFS
+	    for (subvolid_t subvolid : deleted_subvolids)
+	    {
+		while (!does_subvolume_exist(subvolume_dir.fd(), subvolid))
+		    sleep(1);
+	    }
+#endif
 
-	    mnt_table_enable_comments(table, 1);
+	    deleted_subvolids.clear();
+
+	    BtrfsUtils::sync(subvolume_dir.fd());
 	}
+    }
 
-	~MntTable()
-	{
-	    mnt_reset_table(table);
-	}
-
-	void parse_fstab()
-	{
-	    if (mnt_table_parse_fstab(table, target_fstab().c_str()) != 0)
-		throw runtime_error("mnt_table_parse_fstab failed");
-	}
-
-	void replace_file()
-	{
-	    if (mnt_table_replace_file(table, target_fstab().c_str()) != 0)
-		throw runtime_error("mnt_table_replace_file failed");
-	}
-
-	struct libmnt_fs* find_target(const string& path, int directon)
-	{
-	    return mnt_table_find_target(table, path.c_str(), directon);
-	}
-
-	void add_fs(struct libmnt_fs* fs)
-	{
-	    if (mnt_table_add_fs(table, fs) != 0)
-		throw runtime_error("mnt_table_add_fs failed");
-	}
-
-	void remove_fs(struct libmnt_fs* fs)
-	{
-	    if (mnt_table_remove_fs(table, fs) != 0)
-		throw runtime_error("mnt_table_remove_fs failed");
-	}
-
-    private:
-
-	string target_fstab() const
-	{
-	    return prepend_root_prefix(root_prefix, "/etc/fstab");
-	}
-
-	const string root_prefix;
-
-	struct libmnt_table* table;
-
-    };
 
+#ifdef ENABLE_ROLLBACK
 
     void
     Btrfs::addToFstabHelper(const string& default_subvolume_name) const
